/*
 * QEMU CSV support
 *
 * Copyright: Hygon Info Technologies Ltd. 2022
 *
 * Author:
 *      Jiang Xin <jiangxin@hygon.cn>
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 *
 */

#ifndef QEMU_CSV_H
#define QEMU_CSV_H

#include "qapi/qapi-commands-misc-target.h"
#include "qemu/thread.h"
#include "qemu/queue.h"

extern bool kvm_csv_reset_inhibit;

#ifdef CONFIG_SEV

#include "cpu.h"

#define GUEST_POLICY_CSV_BIT     (1 << 6)
#define GUEST_POLICY_REUSE_ASID  (1 << 7)

#define CPUID_VENDOR_HYGON_EBX   0x6f677948  /* "Hygo" */
#define CPUID_VENDOR_HYGON_ECX   0x656e6975  /* "uine" */
#define CPUID_VENDOR_HYGON_EDX   0x6e65476e  /* "nGen" */

static __attribute__((unused)) bool is_hygon_cpu(void)
{
    uint32_t ebx = 0;
    uint32_t ecx = 0;
    uint32_t edx = 0;

    host_cpuid(0, 0, NULL, &ebx, &ecx, &edx);

    if (ebx == CPUID_VENDOR_HYGON_EBX &&
        ecx == CPUID_VENDOR_HYGON_ECX &&
        edx == CPUID_VENDOR_HYGON_EDX)
        return true;
    else
        return false;
}

#else

static __attribute__((unused)) bool is_hygon_cpu(void)
{
    return false;
}

#endif

#ifdef CONFIG_CSV
bool csv_enabled(void);
#else
#define csv_enabled() 0
#endif

struct dma_map_region {
    uint64_t start, size;
    QTAILQ_ENTRY(dma_map_region) list;
};

#define CSV3_OUTGOING_PAGE_WINDOW_SIZE (512 * TARGET_PAGE_SIZE)

extern uint32_t kvm_hygon_coco_ext;
extern uint32_t kvm_hygon_coco_ext_inuse;

struct guest_addr_entry {
    uint64_t share:    1;
    uint64_t reserved: 11;
    uint64_t gfn:      52;
};

struct guest_hva_entry {
    uint64_t  hva;
};

struct CsvGuestState {
    uint32_t policy;
    int sev_fd;
    void *state;
    int (*sev_ioctl)(int fd, int cmd, void *data, int *error);
    const char *(*fw_error_to_str)(int code);
    QTAILQ_HEAD(, dma_map_region) dma_map_regions_list;
    QemuMutex dma_map_regions_list_mutex;
    gchar *send_packet_hdr;
    size_t send_packet_hdr_len;
    struct guest_hva_entry *guest_hva_data;
    struct guest_addr_entry *guest_addr_data;
    size_t guest_addr_len;

    int (*sev_send_start)(QEMUFile *f, uint64_t *bytes_sent);
    int (*sev_receive_start)(QEMUFile *f);
};

typedef struct CsvGuestState CsvGuestState;

extern struct CsvGuestState csv_guest;
extern int csv_init(uint32_t policy, int fd, void *state, struct sev_ops *ops);
extern int csv_launch_encrypt_vmcb(void);
extern struct ConfidentialGuestMemoryEncryptionOps csv_memory_encryption_ops;

int csv_load_data(uint64_t gpa, uint8_t *ptr, uint64_t len, Error **errp);

int csv_shared_region_dma_map(uint64_t start, uint64_t end);
void csv_shared_region_dma_unmap(uint64_t start, uint64_t end);
int csv3_save_outgoing_context(QEMUFile *f);
int csv3_load_incoming_context(QEMUFile *f);
int csv3_load_incoming_page(QEMUFile *f, uint8_t *ptr);
int csv3_queue_outgoing_page(uint8_t *ptr, uint32_t sz, uint64_t addr);
int csv3_save_queued_outgoing_pages(QEMUFile *f, uint64_t *bytes_sent);

int csv3_set_guest_private_memory(Error **errp);

#endif
